/*

Copyright (c) 2009, Barry Watson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Barry Watson ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Barry Watson BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
	
.title "Everyman bootstrap loader"
	 
.eject
.arch	i386

# This is the segment of the bootloader, determined by
# the BIOS which loads us.
BOOT_SEGMENT = 0x7c0

# This is the segment of the disk sector buffer written to by
# the BIOS routine to load sectors.
SECTOR_BUFFER_SEG = 0x400

# This is the segment where we build up a copy of the kernel
# image to be loaded at 1Mbyte. This segment is reachable in
# real mode.
TEMP_KERNEL_SEG	= 0x800

# This is the linear address of the final resting place for
# the kernel. After loading all of the sectors from disk. We
# will copy the image from TEMP_BUFFER_SEG.
KERNEL_START_ADDRESS = 0x100000

# This include file contains KERNEL_SIZE which is the number of
# bytes we should load from disk.
.include "mbr.inc"

# Size of a disk sector.
SECTOR_SIZE = 512

# The number of bytes per paragraph.				
PARAGRAPH_SIZE = 16

# How many sector we will read in from disk.
# The last sector may not be full so we add on
# an extra after dividing.
SECTOR_COUNT = (KERNEL_SIZE/SECTOR_SIZE)+1

# How much we add on to the TEMP_KERNEL_SEG after loading
# a sector.
TEMP_INCREMENT = SECTOR_SIZE/PARAGRAPH_SIZE
	
.code16
.text

.global	_start
_start:
	
# Setup the DS register in case the BIOS did not.
	movw	$BOOT_SEGMENT, %ax		
	movw	%ax, %ds
	
# Save the boot disk id.
	movb	%dl, disk_number	

# Say hello.
	movw	$boot_message, %bp
	movw	$boot_message_len, %cx
	call	write_string
	call	get_disk_geometry
	
# Start loading the kernel image.
load_sector:
# BIOS int 13h, function 02h, Read Sector.
# Input
#	AH = 02h
#	AL = number of sectors to read
#	CH = track/cylinder
#	CL = sector
#	DL = disk (00h-7Fh floppy, 80h-FFh hard disk)
#	DH = head
#	ES:BX = segment:offset of buffer
# Output
#	CF = 0 on success, 1 on failure with AH = failure code
#	AL = number of sectors read
	
	movw	$SECTOR_BUFFER_SEG, %ax		
	movw	%ax, %es
	xorw	%bx, %bx		
	movb	sector_number, %cl	
	movb	cyl_number, %ch		
	movb	disk_number, %dl	
	movb	head_number, %dh	
	movw	$0x0201, %ax
	int	$0x13		
		
	jc	disk_error

	call	copy_sector_to_temp

# Update the counter for the number of sectors left to load.
# Jump to done if the counter is zero.
	decw	sectors_left
	jz	done

# Figure out which sector and head we will use next.
# We increase the head if sector = 64.
	incb	sector_number
	movb	sector_number, %al
	cmpb	$63, %al		
	jle	load_sector
	movb	$1, %al			
	movb	%al, sector_number	
	incb	head_number
	movb	head_number, %dh		
	cmpb    disk_heads, %dh
	jle     1f
	movb    $0, head_number
	incb    cyl_number
	movb	cyl_number, %ch		
	cmpb    disk_cylinders, %ch
	jle     1f
	jmp     disk_error
1:	
	jmp	load_sector

# Tell everyone we have finished loading.
done:	movw	$done_message, %bp
	movw	$done_message_len, %cx		
	call	write_string

# We have to make the code segment base equal to the boot segment.
# In protected mode we can always use absolute data addresses but the
# assembler generates relative addresses for code jumps.
# Register DS holds the initial segment when we were loaded by the BIOS.
	xorl	%ebx, %ebx
	movw	%ds, %bx
	shll	$4, %ebx

# Put the linear address into the descriptor for our code segment	
	movl	%ebx, %eax
	movw	%ax, mbr_code+2
	shrl	$16, %eax
	movb	%al, mbr_code+4
	movb	%ah, mbr_code+7
	
# Calculate the linear address of the Global Descriptor Table. 		
	xorl	%eax, %eax
	movw	$gdt, %ax
	addl	%ebx, %eax
	
# Store it in the proper place for the lgdt instruction.
	movl	%eax, gdtr+2

# Turn on the A20 line.
	cli
	xorw	%cx, %cx
a20_0:	inb	$0x64, %al
	testb	$0x02, %al
	loopnz	a20_0
	movb	$0xd1, %al
	outb	%al, $0x64
	xorw	%cx, %cx
a20_1:	inb	$0x64, %al
	testb	$0x02, %al
	loopnz	a20_1
	movb	$0xdf, %al
	outb	%al, $0x60
	xorw	%cx, %cx
a20_2:	in	$0x64, %al
	testb	$0x02, %al
	loopnz	a20_2

# Move into protected mode.
	lgdt	gdtr	
	movl	%cr0, %eax
 	orb	$1, %al
	movl	%eax, %cr0
	ljmp	$0x10,$protected_mode


# We have had a disk error
disk_error:	
	movw	$fail_message, %bp
	movw	$fail_message_len, %cx
	call	write_string
	
# There is no point carrying on.
loopy:	jmp	loopy

# copy_sector_to_temp
#
# Take the sector which we have just read into memory from the disk
# and copy it to a temporary memory buffer which will contain the
# the whole image after the last sector has been read in and copied

copy_sector_to_temp:	
# Set up the ES segment where we will be copying to.
	movw	temp_ptr, %es
	
# Copy to the beginning of the segment.
	movw	$0, %di			
	
# Set up the DS segment so it points to our
# load buffer segment.
	movw	$SECTOR_BUFFER_SEG, %ax
	movw	%ax, %ds

# Copy from the beginning of the load buffer segment.
	movw	$0, %si			

# Copy 512 bytes.
	movl	$SECTOR_SIZE,%ecx

	cld
	rep
	movsb
	
# Restore the DS segment.
	movw	$BOOT_SEGMENT, %ax
	movw	%ax, %ds
	
# Increase the destination segment with the track size (512) divided by 16.
# As you know segments are multiples of 16.
	addw	$TEMP_INCREMENT, temp_ptr

# That sector is now copied.
	ret

# write_string
#
# Write a string to the screen at position (0,0).
# Pass in bp (pointer to the string) and cx (string length).
	
write_string:
	
# BIOS int 10h, function 13h, Write String
# Input
#	AH = 13h
#	AL = Output mode
#	BH = Display page
#	BL = Attribute
#	CX = Length of string
#	DH = Screen row
#	DL = Screen column
#	ES:BP = Segment:offset of string
# Output
#	Nothing
	
	pushw	%ax
	pushw	%bx
	pushw	%dx
	pushw	%es	
	movw	$0x7c0, %ax	
	movw	%ax, %es	
	xorw	%dx, %dx
	movw	$0x0007, %bx	
	movw	$0x1301, %ax	
	int	$0x10
	popw	%es
	popw	%dx
	popw	%bx
	popw	%ax
	ret
	

# Messages to be written with write_string.

boot_message:	.ascii	"Everyman Bootstrap Loader [ BOOT ]"
boot_message_len	=	.-boot_message

done_message:	.ascii	"Everyman Bootstrap Loader [ DONE ]"
done_message_len	=	.-done_message

fail_message:	.ascii	"Everyman Bootstrap Loader [ FAIL ]"
fail_message_len	=	.-fail_message



get_disk_geometry:	
# BIOS int 13h, function 08h, Get Drive Parameters.
# Input
#	AH = 08h
#	DL = disk (00h-7Fh floppy, 80h-FFh hard disk)
# Output
#	CF = 0 on success, 1 on failure with AH = failure code
#	BL = drive type (01=360K, 02=1.2MB, 03=720K, 04=1.44MB)
#	CH = miximum value of track/cylinder
#	CL = maximum value of sector
#	DH = maximum value of head
#	ES:DI = segment:offset of disk drive paramter table (on failure)
	movb	disk_number, %dl	
	movw	$0x0800, %ax
	int	$0x13			
	jc	disk_error
	movb	%ch, disk_cylinders
	movb	%dh, disk_heads
	ret
	
# Storage space for load information 

disk_number:	.byte   0
sector_number:	.byte   2	# start sector
head_number:	.byte	0	# start header
cyl_number:	.byte	0	# start cylinder
disk_heads:	.byte   0      # see get_disk_geometry
disk_cylinders:	.byte   0	# see get_disk_geometry
sectors_left:	.word   SECTOR_COUNT
temp_ptr:	.word	TEMP_KERNEL_SEG

# This code is executed in protected mode.
.code32
protected_mode:	
	movw	$0x08, %ax
	data16 movw	%ax, %ds
	data16 movw	%ax, %ss
	data16 movw	%ax, %es

# Copy the image from the temporary storage location to 
# the final memory buffer which starts at memory location
# 1 Mbyte.
	
# Compute linear address of the temporary buffer and
# put it in ESI (source).	
	movl	$TEMP_KERNEL_SEG, %eax	
	shll	$4, %eax	
	movl	%eax, %esi

# Put the destination linear address in EDI.
	movl	$KERNEL_START_ADDRESS, %edi

# ECX = length of data to copy in bytes.
	movl	$KERNEL_SIZE, %ecx	

	cld
	rep
	movsb			

# Jump to the kernel with a code segment with offset zero.
# We will never return from this jump.
	ljmp	$0x18,$KERNEL_START_ADDRESS
	


# Global Descriptor Table storage space
.align 4
gdtr:	.word	gdt_end-gdt-1	# number of gdt entries - 1
	.word	0		# linear addresss of table
	.word	0

# Here are the actual descriptors
.align 4
gdt:
# The null descriptor.
	.word	0		
	.word	0
	.word	0
	.word	0
	
# The data descriptor with base zero.
	.word	0xffff
	.word	0
	.byte	0
	.byte	0x92
	.byte	0xcf
	.byte	0

# The MBR code descriptor with non-zero base.	
mbr_code:	
	.word	0xffff
	.word	0
	.byte	0
	.byte	0x9a
	.byte	0xcf
	.byte	0

# The kernel code descriptor with base zero.
	.word	0xffff
	.word	0
	.byte	0
	.byte	0x9a
	.byte	0xcf
	.byte	0

gdt_end	=	.

.org 510
boot_flag:	.word 0xAA55






