/*

Copyright (c) 2009, Barry Watson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Barry Watson ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Barry Watson BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <kernel/kernel.h>
#include <bsp/bsp.h>

extern void system_entrypoint (void);

static sigset_t disable_enable_mask;

/*
 * The simulated hardware clock interrupt service routine.
 * We simply call the KERN_clock_tick function which is basically
 * the system scheduler.
 */
static void clock_signal_handler(int sig)
{
  (void)sig;
  KERN_clock_tick();
}

/*
 * The simulated other hardware interrupt service routine.
 * We call the KERN_job arrive function which moves the job
 * into the pending state so that it can be scheduled.
 *
 * First find the pid of the system.elf program when it is
 * running. Let's say the pid was 7281. We would then give
 * the following command in a shell:
 *  kill -s SIGUSR1 7281
 * The sporadic job would then be scheduled. See app_main.sdl
 * and jobs.c for the definition of this sporadic job.
 */

static void user_signal_handler(int sig)
{
  extern KERN_job_t *s1;

  (void)sig;
  KERN_job_arrive(s1);
}

/*
 * The initialisation routine for the simulated hardware clock.
 * We use a unix interval timer which behaves exactly how a real 
 * hardware clock circuit generating periodic interrupts would behave.
 * We configure the interval timer to give send us an asynchronous signal
 * at 100Hz (every 10 milliseconds). The function clock_signal_handler defined
 * above will be called by the unix OS when the timer expires.
 */
void BSP_init_clock(void)
{
  struct itimerval it;

  it.it_value.tv_sec = 0;
  it.it_value.tv_usec = 10000; /* 100Hz */
  it.it_interval.tv_sec = 0;
  it.it_interval.tv_usec = 10000; /* 100Hz */

  if(setitimer(ITIMER_REAL, &it, NULL) < 0)
    {
      perror("sigaction");
      exit(1);
    }
}

/*
 * This function is called by the everyman kernel when it starts its 
 * initialisation. For the simulated hardware we use unix asynchronous
 * signals instead of interrupts. Therefore, we connect signals and
 * signal handlers here. 
 */
void BSP_init_exception_vectors(void)
{
  struct sigaction sa;

  /*
   * Define the clock signal
   */
  sigemptyset(&disable_enable_mask);
  sigaddset(&disable_enable_mask, SIGALRM);

  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = clock_signal_handler;
  if(sigaction(SIGALRM, &sa, NULL) < 0)
    {
      perror("sigaction");
      exit(1);
    }

  /*
   * Define the other signal connected to
   * the sporadic job.
   */

  sigemptyset(&disable_enable_mask);
  sigaddset(&disable_enable_mask, SIGUSR1);

  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  sa.sa_handler = user_signal_handler;
  if(sigaction(SIGUSR1, &sa, NULL) < 0)
    {
      perror("sigaction");
      exit(1);
    }
 
  /* 
   * No interrupts should be generated until the everyman kernel calls
   * BSP_enable_interrupts for the first time at the end of system initialisation.
   * So we're calling BSP_disable_interrupts for safety's sake
   */
  BSP_disable_interrupts();
}

/* 
 * See BSP_disable_interrupts below. This function just unblocks the previously 
 * blocked signals.
 */
void BSP_enable_interrupts(void)
{
  sigprocmask(SIG_UNBLOCK, &disable_enable_mask, NULL);
}

/*
 * This function is called by the everyman kernel to prevent problems when it
 * uses non reentrant functions. We just block the signals that can lead to
 * kernel code being called to prevent reentrancy problems.
 */
void BSP_disable_interrupts(void)
{
  sigprocmask(SIG_BLOCK, &disable_enable_mask, NULL);
}

/*
 * Somebody has missed their deadline.
 * The everyman kernel calls this function when it discovers a deadline problem.
 *
 * In our case we just exit the program but you would use this function to recover
 * from deadline overruns in a safe way. 
 */
void BSP_deadline_overrun(KERN_job_t *j)
{
  printf("\nError : deadline overrun 0x%x\n", (U32)j);
  exit(1);
}

/*
 * I often use a function called kprintf. This is not part of the BSP, the kernel
 * doesn't need a kprintf function. I often write application code that uses a kprintf
 * function and I define the function here as also I define it in real hardware BSPs.
 * This gives me hardware independent application code. 
 */
void kprintf(char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
}

/*
 * This is where we get the ball rolling. The function KERN_init will 
 * configure the machine using BSP functions, and configure all the jobs,
 * CABs, and resources the application needs. 
 */
int main (void)
{
  KERN_init();

  /* do nothing */
  for(;;);
}
