/*

Copyright (c) 2009, Barry Watson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Barry Watson ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Barry Watson BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <kernel/types.h>
#include <bsp.x86/internal.h>
#include <bsp/bsp.h>

#define	PIC1	   0x20	  /* I/O port address of master 8259a */
#define	PIC2	   0xA0	  /* I/O port address of slave 8258a */
#define	NUM_VEC	   48     /* Number of exceptions */
#define	TRAP       15	  /* Trap Gate */

struct idt 
{
  U16 igd_loffset;
  U16 igd_segsel;
  U32 igd_rsvd : 5;
  U32 igd_mbz : 3;
  U32 igd_type : 5;
  U32 igd_dpl : 2;
  U32 igd_present : 1;
  U16 igd_hoffset;
};

struct trapframe
{
  U32 ebp_extra;
  U32 edi;
  U32 esi;
  U32 ebp;
  U32 esp;
  U32 ebx;
  U32 edx;
  U32 ecx;
  U32 eax;
  U32 trap_number;
  U32 code;
  U32 eip;
  U32 ecs;
  U32 eflags;
};


/*
 * The interrupt_mask is used when we enable interrupts with
 * BSP_enable_interrupts() which is defined in locore.S
 */
U16 interrupt_mask;

/*
 * The default exception table which we will fill with the 
 * default handler.
 */
extern U32	default_vec[];

/* The following are defined in locore.S */
extern struct idt idt[NUM_VEC];
extern struct segtr idtr;

/* Used to save the stack pointer when we look at the trap frame */
static  U8 * stack_pointer;

void set_vec(U32 exception_number, U32 handler);
static void init_8259a (void);

void BSP_init_exception_vectors(void)
{
  U32 i;
  
  for (i=0; i<NUM_VEC; ++i)
    {
      set_vec(i, default_vec[i]);
    }

  interrupt_mask = 0xfffb;	/* leave bit 2 enabled for IC cascade */
  
  lidt();
  init_8259a();
}

void set_vec(U32 exception_number, U32 handler)
{
  struct idt *pidt;
  
  pidt = &idt[exception_number];
  pidt->igd_loffset = handler;
  pidt->igd_segsel = 0x8;		/* kernel code segment */
  pidt->igd_mbz = 0;
  pidt->igd_type = TRAP;
  pidt->igd_dpl = 0;
  pidt->igd_present = 1;
  pidt->igd_hoffset = handler >> 16;
  
  
  if (exception_number > 31 && exception_number < 48) 
    {
      /* 
       * This is a bus interrupt so enable the bit in the interrupt_mask. 
       * Actually, we clear the bit, 1 => disable, 0 => enable
       */
      exception_number = exception_number - 32;
      interrupt_mask &= ~(1 << exception_number);
    }
}

U8 *trap_description[19] = 
{
  "Divide error",
  "Single-step",
  "NMI",
  "Breakpoint (INT3)",
  "Overflow",
  "BOUND",
  "Invalid opcode",
  "Coprocessor not available",
  "Double fault",
  "Coprocessor segment overrun",
  "Invalid TSS",
  "Segment not present",
  "Stack segment overrun",
  "General protection",
  "Page fault",
  "Unassigned (Type 15)",
  "Coprocessor error",
  "Alignment check",
  "Machine check"
};

void trap(void)
{
  struct trapframe *tfp;
  
  BSP_disable_interrupts();
  
  /* 
   * Get our hands on the stack pointer
   */
  asm("movl  %ebp, stack_pointer");
  
  /* 
   * and make its contents look friendly
   */
  tfp = (struct trapframe *)stack_pointer;
  
  
  /* 
   * Now we just display the contents of the frame
   * for Barry so he can get started with debugging.
   */
  kprintf("Exception no. %x with extra data 0x%x\n", 
	  tfp->trap_number,
	  tfp->code);
  
  if(tfp->trap_number <= 18)
    {
      kprintf("%s\n", trap_description[tfp->trap_number]);
    }
  else if(tfp->trap_number <= 31)
    {
      kprintf("Reserved trap number\n");
    }
  else
    {
      kprintf("Peripheral device interrupt. Is your driver installed?\n");
    }
  kprintf("EIP = 0x%x\n", tfp->eip);
  kprintf("ECS = 0x%x\n", tfp->ecs);  
  kprintf("EFLAGS = 0x%x\n", tfp->eflags);
  kprintf("EAX = 0x%x\n", tfp->eax);
  kprintf("EBX = 0x%x\n", tfp->ebx);
  kprintf("ECX = 0x%x\n", tfp->ecx);
  kprintf("EDX = 0x%x\n", tfp->edx);
  kprintf("EDI = 0x%x\n", tfp->edi);
  kprintf("ESI = 0x%x\n", tfp->esi);
  kprintf("EBP = 0x%x\n", tfp->ebp);
  kprintf("ESP = 0x%x\n", tfp->esp);

  /*
   * If we ever execute this code then it is bad news.
   * I don't reckon we want to return from the exception.
   * Remember interrupts are disabled before we execute this
   * for loop, what is going to preempt us?
   */ 
  for(;;);
}


static void init_8259a (void)
{ 
  /****************/
  /* master 8259a */
  /****************/

  /*************************/  
  /* icw4 needed           */
  /* cascade mode          */
  /* address interval of 8 */
  /* edge triggered mode   */
  /*************************/
  outb(PIC1, 0x11);
  
  /******************/
  /* base irq is 32 */
  /******************/
  outb(PIC1+1, 0x20);	

  /**********************************/
  /* IR input 2 (bit 3) has a slave */
  /**********************************/
  outb(PIC1+1, 0x4);

  /*********************************/
  /* 8086/8088 mode                */
  /* normal EOI                    */
  /* non buffered mode             */
  /* not special fully nested mode */
  /*********************************/
  outb(PIC1+1, 0x1);

  /*********************************/
  /* read IS reg on next RD pulse  */
  /* no poll                       */
  /* no action (special mask mode) */
  /*********************************/
  outb(PIC1, 0xb);	
  
  /****************/
  /* slave 80259a */ 
  /****************/

  /*************************/
  /* icw4 needed           */
  /* cascade mode          */
  /* address interval of 8 */
  /* edge triggered mode   */
  /*************************/
  outb(PIC2, 0x11);	

  /******************/
  /* base irq is 40 */
  /******************/
  outb(PIC2+1, 0x28);

  /***********************/
  /* slave identity is 2 */
  /***********************/
  outb(PIC2+1, 0x2);

  /*****************************/
  /* 8086/8088 mode            */
  /* auto EOI                  */
  /* non buffered mode         */
  /* special fully nested mode */
  /*****************************/
  outb(PIC2+1, 0xb);

  /*********************************/
  /* read IS reg on next RD pulse  */
  /* no poll                       */
  /* no action (special mask mode) */
  /*********************************/
  outb(PIC2, 0xb);
  
  BSP_disable_interrupts();
}


