/*

Copyright (c) 2009, Barry Watson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Barry Watson ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Barry Watson BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <bsp/bsp.h>

#define NULL 0

static KERN_job_t the_jobs[MAX_NUMBER_JOBS];
static U32 job_index;
static KERN_resource_t the_resources[MAX_NUMBER_RESOURCES];
static U32 resource_index;
static void *stack[MAX_STACK_DEPTH];
static U32 stack_index;
static U32 system_ceiling;
static KERN_job_t *current_job;        /* The job (periodic or sporadic) that is currently executing */
static KERN_job_t *pending_jobs_list;  /* The periodic jobs that want to execute */
static KERN_job_t *inactive_jobs_list; /* The periodic jobs that await their next period */
static U32 system_time;

static void job_preempt (void);
static void job_add_to_inactive_jobs_list (KERN_job_t *job);
static void job_arrive (KERN_job_t *job);


/*
 * This function is reentrant. Obvious eh?
 *
 * We call this function when we want to update the kernel's internal
 * data structures. 
 */
static void system_lock (void)
{
    BSP_disable_interrupts();
}

/*
 * This function is reentrant.
 *
 * We call this function when we've finished updating the kernel's 
 * internal data structures.
 */
static void system_unlock (void)
{
    BSP_enable_interrupts();
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 *
 * This function is called to save system state we need to "remember"
 * when we preempt jobs or perform resource requests.  
 */
static void stack_push (void *data)
{
    stack[stack_index++] = data;
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 *
 * This function is called to restore a previous system state that
 * was saved with "stack_push". We retrieve the data when we finish
 * a preemption (current job finishes execution) and when resources
 * are released
 */
static void *stack_pop (void)
{
    return stack[--stack_index];
}

/*
 * This function is reentrant.
 */
void KERN_resource_request (KERN_resource_t *res, U32 amount)
{
    U32 resource_ceiling;

    /* 
     * It's possible that buggy code would try to request more than is available,
     * we really should check that here, otherwise, we're going to get a problem.
     * We need some sort of mechanism which prevents this and stops unbalanced
     * releases....... hmmmmmm.........
     */

    system_lock();

    stack_push((void *)system_ceiling); /* The system ceiling may be raised later */
    stack_push((void *)res->available); /* The amount available will be lowered */    
    stack_push(res);                    /* So we know which resource to release */

    res->available = res->available - amount;
    resource_ceiling = res->ceiling[res->available];
    if(resource_ceiling > system_ceiling)
    {
	system_ceiling = resource_ceiling;
    }

    system_unlock();
}

/*
 * This function is reentrant.
 */
void KERN_resource_release (void)
{
    KERN_resource_t *res;
    U32 old_system_ceiling;

    system_lock();

    res = stack_pop(); /* Get the last resource requested but not released */
    res->available = (U32)stack_pop();
    old_system_ceiling = system_ceiling;
    system_ceiling = (U32)stack_pop();

    if(system_ceiling < old_system_ceiling)
    {
	system_unlock();
	job_preempt();
    }
    else
    {
	system_unlock();
    }
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 */
KERN_resource_t *KERN_resource_create(U32 available, U32 *ceiling)
{
  KERN_resource_t *res;

  if(resource_index > MAX_NUMBER_RESOURCES - 1)
    {
      return NULL;
    }
  
  res = &the_resources[resource_index++];
  res->available = available;
  res->ceiling = ceiling;
  
  return res;
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 */
KERN_job_t *KERN_job_create (U32 period, U32 relative_deadline, void (*function)(void), KERN_job_type_t type)
{
  KERN_job_t *job;

  if(job_index > MAX_NUMBER_JOBS - 1)
    {
      return NULL;
    }
  
  job = &the_jobs[job_index++];
  job->type = type;
  job->period = period;
  job->relative_deadline = relative_deadline;
  job->function = function;
  job->preemption_level = 0xffffffff - relative_deadline; /* 32-bit specific */

  if(job->type == PERIODIC_JOB)
    {      
      job->absolute_deadline = system_time + job->relative_deadline;
      job_arrive(job);
    }
  return job;
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 */
static void job_add_to_pending_jobs_list (KERN_job_t *job)
{
    KERN_job_t *runner;
    KERN_job_t *last_runner;

    for(last_runner=NULL, runner = pending_jobs_list; runner; last_runner=runner, runner = runner->next_job)
    {
#if defined OPT_EDF_SCHEDULING
	if(runner->absolute_deadline > job->absolute_deadline)
#elif defined OPT_DM_SCHEDULING
	if(runner->relative_deadline > job->relative_deadline)
#else
#error "no scheduler defined"
#endif
	{
	    break;
	}
    }
    if(runner)
    {
	job->prev_job = runner->prev_job;	
	if(runner->prev_job)
	{
	    runner->prev_job->next_job = job; 
	}
	else
	{
	    pending_jobs_list = job;
	}

	job->next_job = runner;
	runner->prev_job = job;

    }
    else
    {
	if(last_runner)
	{
	    last_runner->next_job = job;
	    job->prev_job = last_runner;
	    job->next_job = NULL;
	}
	else
	{
	    job->prev_job = job->next_job = NULL;
	    pending_jobs_list = job;
	}
    }
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 *
 * This function is called when either:
 * - A periodic job's period has started.
 * - A sporadic job has been released.
 */
static void job_arrive (KERN_job_t *job)
{
  if(job->type == SPORADIC_JOB)
    {
      job->absolute_deadline = system_time + job->relative_deadline;
    }
  job_add_to_pending_jobs_list(job);

#ifdef OPT_USE_HOOKS
    HOOK_job_arrive(job);
#endif
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 *
 * See the function "job_finish".
 * This function is called when a job has finished execution and 
 * we want it to wait until the start of its next period.
 * This function should not be used with sporadic jobs.
 */
static void job_add_to_inactive_jobs_list (KERN_job_t *job)
{
    KERN_job_t *runner;
    KERN_job_t *last_runner;

    for(last_runner=NULL, runner = inactive_jobs_list; runner; last_runner = runner, runner = runner->next_job)
    {
	if(runner->arrival_time > job->arrival_time)
	{
	    break;
	}
    }
    if(runner)
    {
	job->prev_job = runner->prev_job;	
	if(runner->prev_job)
	{
	    runner->prev_job->next_job = job; 
	}
	else
	{
	    inactive_jobs_list = job;
	}

	job->next_job = runner;
	runner->prev_job = job;

    }
    else
    {
	if(last_runner)
	{
	    last_runner->next_job = job;
	    job->prev_job = last_runner;
	    job->next_job = NULL;
	}
	else
	{
	    job->prev_job = job->next_job = NULL;
	    inactive_jobs_list = job;
	}
    }
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 *
 * This function is called when a job finishes execution. 
 * We set up the job's timing parameters for the next period.
 */
static void job_finish (KERN_job_t *job)
{
  if(job->type == PERIODIC_JOB)
    {
      job->arrival_time = (job->arrival_time + job->period);
      job->absolute_deadline = job->arrival_time + job->relative_deadline;
      job_add_to_inactive_jobs_list(job);
    }
#ifdef OPT_USE_HOOKS
    HOOK_job_finish(job);
#endif

    current_job = stack_pop(); /* pop the current job's id from the stack */
}

/*
 * This function is not reentrant. It should be called with interrupts disabled.
 *
 * Remove the job at the head of the "job_list".
 * This function is currently only used on the pending jobs list.
 */
static KERN_job_t *job_list_unlink (KERN_job_t **job_list)
{
    KERN_job_t *retval;

    if(*job_list)
    {
	retval = *job_list;
	if(retval->next_job)
	{
	    retval->next_job->prev_job = NULL;
	}
	*job_list = retval->next_job;
    }
    else
    {
	retval = NULL;
    }

    return retval;
}

/*
 * This function is reentrant.
 *
 * This function is called to check for any possible preemption.
 * The purpose is basically to find the job with the highest priority 
 * which is either executing, or, pending and not blocked by the preemption level.
 * If this job isn't the currently executing job then we make sure that it becomes
 * the currently executing job. 
 */
static void job_preempt (void)
{
  system_lock();
  for(;;)
    {
      if(pending_jobs_list)
	{
	  if(!current_job || 
#if defined OPT_EDF_SCHEDULING
	     ((current_job->absolute_deadline > pending_jobs_list->absolute_deadline) &&
#elif defined OPT_DM_SCHEDULING
	     ((current_job->relative_deadline > pending_jobs_list->relative_deadline) &&
#else
#error "no scheduler defined"
#endif
	      (system_ceiling < pending_jobs_list->preemption_level)))
	    {
	      /* We need to preempt the currently executing job */

	      /* Remember the current job so we can return to executing it later */
	      stack_push(current_job);

	      /* Get the new current job */
	      current_job = job_list_unlink(&pending_jobs_list);
	      
#ifdef OPT_USE_HOOKS
	      HOOK_job_start(current_job);
#endif
	      system_unlock();

	      /* Start executing the new current job */
	      current_job->function();

	      system_lock();
	      job_finish(current_job);
	    }
	  else
	    {
	      /* No preemption - the current job has the highest priority.*/
	      break;
	    }
	}
      else
	{
	  /* No preemption - there are no pending jobs */
	  break;
	}
    } /* for(;;) */
  system_unlock();
}

/*
 * This function is reentrant.
 *
 * This function is called at every system clock tick. 
 * This is the core of system scheduling.
 * The "system lock" function prevents function execution. 
 */
void KERN_clock_tick (void)
{
  KERN_job_t *j, *temp;
  
  system_lock();
  system_time++;
  
  /* Has the current job missed its deadline */
  if(current_job && system_time == current_job->absolute_deadline)
    {
      BSP_deadline_overrun(current_job);
    }

  /* Have any pending jobs missed their deadlines */
  for(j = pending_jobs_list; j; j=j->next_job)
    {
      if(system_time == j->absolute_deadline)
	{
	  BSP_deadline_overrun(j);
	}
    }

  /* See if any jobs have started their periods */
  for(j = inactive_jobs_list; j;)
  {
      if(system_time == j->arrival_time)
      {
	  temp = j;
	  j=j->next_job;
	  if(temp == inactive_jobs_list)
	  {
	      inactive_jobs_list = j;
	      if(j)
	      {
		  j->prev_job = NULL;
	      }
	  }
	  else        /* shouldn't need this "else" */
	  {
	      temp->prev_job->next_job = j;
	      if(j)
	      {
		  j->prev_job = temp->prev_job;
	      }
	  }
	  /* Move the job into the pending jobs list */
	  job_arrive(temp);
      }
      else
      {
	  j=j->next_job;
      }
  }
  system_unlock();

  /* We may have a possible preemption */
  job_preempt();
}

/*
 * This function is reentrant.
 *
 * Don't call this function on periodic jobs or when the interrupts
 * are disabled.
 *
 * This function is called in order to get a sporadic job into the pending
 * jobs list.  
 *
 * If you really need move a job into the pending state and interrupts are disabled, 
 * then use the internal function "job_arrive" instead and when you enable interrupts
 * again call "job_preempt". 
 */
void KERN_job_arrive (KERN_job_t *job)
{
  system_lock();
  job_arrive(job);
  system_unlock();
  job_preempt();
}

/*
 * This function is reentrant.
 */
void KERN_cab_create(KERN_cab_t *the_cab, U8 *memory, U32 number_of_buffers, U32 buffer_size)
{
  KERN_cab_buffer_t *acab;
  U8 *runner;
 
  the_cab->free_list = (KERN_cab_buffer_t *)memory;
  the_cab->number_of_buffers = number_of_buffers;
  the_cab->most_recent = NULL;

  for(runner=memory; 
      runner < memory+(number_of_buffers-1)*(sizeof(KERN_cab_buffer_t) + buffer_size - 1); 
      runner += sizeof(KERN_cab_buffer_t) + buffer_size - 1)
  {
      acab = (KERN_cab_buffer_t *)runner;
      acab->next = (KERN_cab_buffer_t *)(runner + sizeof(KERN_cab_buffer_t) + buffer_size - 1);
      acab->use_counter = 0;
  }
  acab = (KERN_cab_buffer_t *)runner;
  acab->next = NULL;
  acab->use_counter = 0;
  
}

/*
 * This function is reentrant.
 */
KERN_cab_buffer_t * KERN_cab_reserve(KERN_cab_t *cab)
{
  KERN_cab_buffer_t *buffer;

  system_lock();
  buffer = cab->free_list;
  if(!buffer)
    {
      /* This is due to the cab being wrongly dimensioned */
      goto out;
    }
  cab->free_list = cab->free_list->next;
 out:
  system_unlock();
  return buffer;
}

/*
 * This function is reentrant.
 */
KERN_cab_buffer_t * KERN_cab_get(KERN_cab_t *cab)
{
  KERN_cab_buffer_t *buffer;
  
  system_lock();
  buffer = cab->most_recent;
  if(!buffer)
    {
      goto out;
    }
  buffer->use_counter++;
 out:
  system_unlock();
  return buffer;
}

/*
 * This function is reentrant.
 */
void KERN_cab_put(KERN_cab_t *cab, KERN_cab_buffer_t *buffer)
{
    system_lock();
    if(cab->most_recent)
    {
	if(!cab->most_recent->use_counter)
	{
	  /* 
	   * Since no jobs are using the previously "most recent" buffer, 
	   * we free it.
	   */
	    cab->most_recent->next = cab->free_list;
	    cab->free_list = cab->most_recent; 
	}
    }
    cab->most_recent = buffer;
    system_unlock();
}

/*
 * This function is reentrant.
 */
void KERN_cab_unget(KERN_cab_t *cab, KERN_cab_buffer_t *buffer)
{
    system_lock();
    if(!--buffer->use_counter && buffer != cab->most_recent)
    {
	buffer->next = cab->free_list->next;
	cab->free_list = buffer;	
    }  
    system_unlock();
}

/*
 * This function is reentrant.
 */
void * KERN_cab_buffer_data (KERN_cab_buffer_t *buffer)
{
  return (void *)buffer->data;
}




