/*

Copyright (c) 2009, Barry Watson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Barry Watson ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Barry Watson BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdlib.h>
#include "hash_table.h"

struct hash_table
{
  struct hash_table *next;
  char *name;
  void *payload;
};


/*
 * Calculate the hash value of a given string.
 * This is from Aho, Sethi, and Ullman.
 */

static int hash (char * s)
{
  char *p;
  unsigned h = 0, g;

  for(p = s; *p; p++)
    {
      h = (h << 4) + *p;
      if(g = h&0xf0000000)
	{
	  h = h ^ ( g >> 24);
	  h = h ^ g;
	}
    }
  return h % HASH_TABLE_SIZE;
}


/*
 * Create a hash_table
 * Returns NULL on failure.
 */

hash_table_t hash_table_create (void)
{
  return malloc(sizeof(struct hash_table *) * HASH_TABLE_SIZE);
}



/*
 * Lookup a given name in a hash table.
 * Returns NULL on failure. 
 * Otherwise returns the payload.
 */

void * hash_table_lookup (hash_table_t table, char *s)
{
    struct hash_table * runner;
    for(runner = table[hash(s)]; runner; runner = runner->next)
    {
	if(!strcmp(runner->name, s))
	{
	    return runner->payload;
	}
    }
    return NULL;
}

/*
 * Put data into the hash table. 
 * "s" is the name and "p" is the payload.
 * We return 1 on success and 0 on failure.
 */

int hash_table_insert (hash_table_t table, char *s, void *p)
{
  int hashval;
  struct hash_table * runner;
  hashval = hash(s);
  for(runner = table[hashval]; runner; runner = runner->next)
    {
      if(!strcmp(runner->name, s))
      {
	  runner->payload = p;
	  return 1;
      }
    }
  
  runner = malloc(sizeof(struct hash_table));
  if(!runner)
    {
      return 0;
    }
  runner->next = table[hashval];
  runner->name = s;
  runner->payload = p;
  table[hashval] = runner;
  
  return 1;
}

/*
 * Apply the function "f" to every data item stored in the
 * hash table.
 */

void hash_table_map(hash_table_t table, void (*f)(void *))
{
  struct hash_table * runner;
  int i;

  for(i=0; i<HASH_TABLE_SIZE; i++)
  {
      for(runner = table[i]; runner; runner = runner->next)
      {
	  (*f)(runner->payload);
      }
  }
}
