%{
/*

Copyright (c) 2009, Barry Watson
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Barry Watson ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Barry Watson BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include <stdlib.h>	
#include <stdio.h>
#include <string.h>

#include "hash_table.h"
#include "kernel/kernel.h"

#define PREEMPTION_LEVEL(d) (0xffffffff - (d)) /* 32-bit specific */
    
extern int linecount;
static FILE *yyout_c;
static FILE *yyout_h;
extern FILE *yyin;

static char *infilename;
static hash_table_t symbol_table;
 
static int job_count;
static int resource_count;
 
struct use_list
{
    unsigned int amount;
    char *id;
    struct use_list *next;
};

static struct use_list *make_use_list_element(unsigned int amount, char *id);
static struct use_list *link_in_use_list_element(struct use_list *elmt, struct use_list *lst);

/*
 * This is the datatype that we use for the hash table payload.
 * See the hash_table.c file.
 */
typedef enum {kPERIODIC, kSPORADIC} job_type_t; 
struct table_entry
{
    enum {kRESOURCE, kJOB, kMACRO, kCAB} kind;
    union
    {
	struct
	{
	    char *id;
	    unsigned int count;
	    unsigned int *ceiling_table;
	  unsigned int available;
	} resource;
	
	struct
	{
	    job_type_t type;
	    char *id;
	    unsigned int deadline;
	    unsigned int period;
	    char *entrypoint;
	    struct use_list *use_list;
	} job;	

	struct
	{
	    char *id;
	    struct use_list *use_list;
	} macro;

	struct
	{
	    char *id;
	    unsigned int number;
	    unsigned int count;
	} cab;
    } u;
};

enum {kUNDEFINED, kEDF, kDM} scheduler = kUNDEFINED;

static void make_resource(char *id, unsigned int count, unsigned int available);
static void make_job(char *id, unsigned int deadline, unsigned int period, char *entrypoint, job_type_t type, struct use_list *use_list);
static void make_macro_definition(char *id, struct use_list *use_list);
static struct use_list *get_macro_definition(char *id);
static struct use_list *normalise_use_list(struct use_list *use_list);
static void make_cab(char *id, unsigned int count);
static void parse_option(char *id);
static void pass0(void *);
static void pass1(void *);
static void pass2(void *);
static void pass3(void *);

%}	

%union
{
    char * str;
    unsigned int num;
    struct use_list *ul;
}
%token PERIODIC SPORADIC DEADLINE PERIOD ENTRYPOINT RESOURCE COUNT USES OF AVAILABLE OPTION MACRO CAB
%token <str>ID 
%token <num>NUM
%type <ul>use_definition
%type <ul>use_definition_list
%%

start: options resources cabs macros jobs
;

options: 
| options_list;

options_list: option
| options_list option
;

option: OPTION ID {parse_option($2);}
;

resources:
| resources_list
;

resources_list: resource
| resources_list resource
;

resource: RESOURCE ID COUNT NUM AVAILABLE NUM                             {make_resource($2, $4, $6);}
;

cabs:
| cabs_list
;

cabs_list: cab
| cabs_list cab
;

cab: CAB ID COUNT NUM {make_cab($2, $4);}
;

macros:
| macros_list
;

macros_list: macro
| macros_list macro
;

macro: MACRO ID use_definition_list {make_macro_definition($2, $3);}
;

jobs:
| jobs_list
;

jobs_list: job
| jobs_list job
;

job: PERIODIC ID DEADLINE NUM PERIOD NUM ENTRYPOINT ID                            {make_job($2, $4, $6, $8, kPERIODIC, NULL);} 
| PERIODIC ID DEADLINE NUM PERIOD NUM ENTRYPOINT ID use_definition_list           {make_job($2, $4, $6, $8, kPERIODIC, $9);} 
| SPORADIC ID DEADLINE NUM PERIOD NUM ENTRYPOINT ID                               {make_job($2, $4, $6, $8, kSPORADIC, NULL);} 
| SPORADIC ID DEADLINE NUM PERIOD NUM ENTRYPOINT ID use_definition_list           {make_job($2, $4, $6, $8, kSPORADIC, $9);} 
;

use_definition_list: use_definition {$$ = $1;}
|  use_definition_list use_definition {$$ = link_in_use_list_element($2, $1);}
;

use_definition: USES NUM OF ID {$$ = make_use_list_element($2, $4);}
| USES ID {$$ = get_macro_definition($2);}
;


%%

void usage (char *name)
{
  fprintf(stderr, "Usage: %s kernel-skeleton-file input-file\n", name);   
  exit(1);	    
}

int main(int argc, char **argv)
{
    FILE *kernel_code;
    int ch;

    printf("\nSDL\n");
    printf("The Everyman kernel's System Description Language compiler\n");

    yyout_c = fopen("sdl.out.c", "w");
    
    if(!yyout_c)
    {
	fprintf(stderr, "ERROR: Could not open 'sdl.out.c'\n");
	exit(1);
    }
    
    yyout_h = fopen("sdl.out.h", "w");
    
    if(!yyout_h)
    {
	fprintf(stderr, "ERROR: Could not open 'sdl.out.h'\n");
	exit(1);
    }

    
    if(argc == 3)
    {

	kernel_code = fopen(argv[1], "r");
	
	if(!kernel_code)
	{
	    fprintf(stderr, "ERROR: Could not open kernel code include file '%s'\n", argv[1]);
	    exit(1);
	}
	
	yyin = fopen(argv[2], "r");
	
	if(!yyin)
	{
	    fprintf(stderr, "ERROR: Could not open SDL code input file '%s'\n", argv[2]);
	    exit(1);
	}
	infilename = argv[2];
    }
    else
    {
	usage(argv[0]);
    }
    

    symbol_table = hash_table_create();
    yyparse();
    

    fprintf(yyout_h, "#include <kernel/kernel.h>\n");

    hash_table_map(symbol_table, pass0);
    fprintf(yyout_c, "#define MAX_NUMBER_JOBS %d\n", job_count);
    fprintf(yyout_c, "#define MAX_NUMBER_RESOURCES %d\n", resource_count);

    /* 
     * You may want to change the calculation of MAX_STACK_DEPTH. 
     * See Step 2 of the book for details. 
     */
    fprintf(yyout_c, "#define MAX_STACK_DEPTH %d\n", 3*resource_count+job_count); 
    
    while((ch = fgetc(kernel_code)) != EOF)
    {
	fputc(ch, yyout_c);
    }

    hash_table_map(symbol_table, pass1);
    hash_table_map(symbol_table, pass2);
    fprintf(yyout_c, "\nvoid KERN_init(void)\n{\n");
    hash_table_map(symbol_table, pass3);
    fprintf(yyout_c, "\tBSP_init_exception_vectors();\n");
    fprintf(yyout_c, "\tBSP_init_clock();\n");
    fprintf(yyout_c, "\tsystem_unlock();\n");
    fprintf(yyout_c, "\tjob_preempt();\n}\n");

    fclose(yyout_c); /* this'll be closed anyway when we exit */
    fclose(yyout_h); /* this'll be closed anyway when we exit */
    fclose(kernel_code); /* this'll be closed anyway when we exit */

    exit(0);
}


static void *safe_malloc(size_t size)
{
    void *retval;
    
    retval = malloc(size);
    if(!retval)
    {
	fprintf(stderr, "Error: malloc returned NULL, out of memory\n");
	exit(1);
    }
    return retval;
}

static void make_resource(char *id, unsigned int count, unsigned int available)
{
    struct table_entry *res;
    
    if(available > count)
    {
	fprintf(stderr, "Error: line %d: available greater than count\n", linecount);
	exit(1);
    }
    
    res = (struct table_entry *)safe_malloc(sizeof (struct table_entry));
    res->kind = kRESOURCE;
    res->u.resource.id = id;
    res->u.resource.count = count;
    res->u.resource.available = available;
    /* here we see (1 + count) as we must include space for zero */
    res->u.resource.ceiling_table = (unsigned int *)safe_malloc((1 + count) * sizeof(unsigned int));
    memset(res->u.resource.ceiling_table, 0, (1 + count) * sizeof(unsigned int));
    
    if(hash_table_lookup(symbol_table, id))
    {
	fprintf(stderr, "Error: line %d: The resource id %s is multiply defined\n", linecount, id);
	exit(1);
    }
    if(!hash_table_insert(symbol_table, id, res))
    {
	fprintf(stderr, "Error: Could not make the entry in the hash table\n");
	exit(1);
    }
}

static void make_job(char *id, unsigned int deadline, unsigned int period, char *entrypoint, job_type_t type, struct use_list *use_list)
{
    struct table_entry *job;
    
    job = (struct table_entry *)safe_malloc(sizeof (struct table_entry));
    job->kind = kJOB;
    job->u.job.type = type;
    job->u.job.id = id;
    job->u.job.deadline = deadline;
    job->u.job.period = period;
    job->u.job.entrypoint = entrypoint;
    job->u.job.use_list = normalise_use_list(use_list);
    
    if(hash_table_lookup(symbol_table, id))
    {
	fprintf(stderr, "Error: line %d: The job id %s is multiply defined\n", linecount, id);
	exit(1);
    }
    if(!hash_table_insert(symbol_table, id, job))
    {
	fprintf(stderr, "Error: Could not make the entry in the hash table\n");
	exit(1);
    }
}

static void parse_option(char *id)
{
    if(!strcmp(id, "hooks"))
    {
	fprintf(yyout_c, "#define OPT_USE_HOOKS\n");
    }
    else if(!strcmp(id, "dm"))
    {
	fprintf(yyout_c, "#define OPT_DM_SCHEDULING\n");
	scheduler = kDM;
    }
    else if(!strcmp(id, "edf"))
    {
	fprintf(yyout_c, "#define OPT_EDF_SCHEDULING\n");
	scheduler = kEDF;
    }
    else
    {
	fprintf(stderr, "Error: The option \"%s\" is not a valid option.\n", id);
	exit(1);
    }
}

static struct use_list *make_use_list_element(unsigned int amount, char *id)
{
    struct use_list *retval;
    
    retval = (struct use_list *)safe_malloc(sizeof(struct use_list));
    retval->amount = amount;
    retval->id = id;
    retval->next = NULL;
    
    return retval;
}

static struct use_list *link_in_use_list_element(struct use_list *elmt, struct use_list *lst)
{
    elmt->next = lst;
    return elmt;
}

static void make_macro_definition(char *id, struct use_list *use_list)
{
    struct table_entry *macro;
    
    macro = (struct table_entry *)safe_malloc(sizeof (struct table_entry));
    macro->kind = kMACRO;
    macro->u.macro.id = id;
    macro->u.macro.use_list = use_list;

    if(hash_table_lookup(symbol_table, id))
    {
	fprintf(stderr, "Error: line %d: The macro id %s is multiply defined\n", linecount, id);
	exit(1);
    }
    if(!hash_table_insert(symbol_table, id, macro))
    {
	fprintf(stderr, "Error: Could not make the entry in the hash table\n");
	exit(1);
    }
}

static struct use_list *get_macro_definition(char *id)
{
    struct table_entry *te;

    if((te = hash_table_lookup(symbol_table, id)) == NULL)
    {
	fprintf(stderr, "Error: line %d: The id %s is not a defined macro\n", linecount, id);
	exit(1);
    }
    if(te->kind != kMACRO)
    {
	fprintf(stderr, "Error: line %d: The id %s is not a defined macro\n", linecount, id);
	exit(1);
    }
    return te->u.macro.use_list;
}

static struct use_list *normalise_insert(struct use_list *elmt, struct use_list *lst)
{
    struct use_list *runner;

    for(runner=lst; runner; runner=runner->next)
    {
	if(!strcmp(runner->id, elmt->id))
	{
	    runner->amount += elmt->amount;
	    return lst;
	}
    }
    elmt->next=lst;
    return elmt;
}
 
static struct use_list *normalise_use_list(struct use_list *use_list)
{
    if(use_list == NULL)
    {
	return NULL;
    }
    return normalise_insert(use_list, normalise_use_list(use_list->next));
}

static void make_cab(char *id, unsigned int count)
{
    struct table_entry *res;
    
    res = (struct table_entry *)safe_malloc(sizeof (struct table_entry));
    res->kind = kCAB;
    res->u.cab.id = id;
    res->u.cab.number = 0;
    res->u.cab.count = count;
    if(hash_table_lookup(symbol_table, id))
    {
	fprintf(stderr, "Error: line %d: The cab id %s is multiply defined\n", linecount, id);
	exit(1);
    }
    if(!hash_table_insert(symbol_table, id, res))
    {
	fprintf(stderr, "Error: Could not make the entry in the hash table for the cab %s\n", id);
	exit(1);
    }
}

/*
 * Count the number of jobs and resources defined.
 */
static void pass0(void *table_entry)
{
    struct table_entry *te;
    
    te = (struct table_entry *)table_entry;
    if(te->kind == kJOB)
    {
	job_count++;
    }
    else if(te->kind == kRESOURCE)
    {
	resource_count++;
    }
}

/*
 * Configure the CABs (how many concurrent users) and resources (ceiling table).
 */
static void pass1(void *table_entry)
{
    struct table_entry *te;
    struct use_list *elmt;
    struct table_entry *resource;
    unsigned int i;
    
    te = (struct table_entry *)table_entry;
    if(te->kind == kJOB)
    {
	
	for(elmt = te->u.job.use_list; elmt; elmt = elmt->next)
	{
	    resource = hash_table_lookup(symbol_table, elmt->id);
	    if(!resource)
	    {
		fprintf(stderr, "Error: no such resource or cab %s\n", elmt->id);
		exit(1);
	    }
	    if(resource->kind == kRESOURCE)
	    {
		if(resource->u.resource.count < elmt->amount)
		{
		    fprintf(stderr, "Error: Use count exceeded\n");
		    exit(1);
		}
		for(i=0; i<elmt->amount; i++)
		  {
		    if(resource->u.resource.ceiling_table[i] < PREEMPTION_LEVEL(te->u.job.deadline))
		      {
			resource->u.resource.ceiling_table[i] = PREEMPTION_LEVEL(te->u.job.deadline);
		      }
		  }				
	    }
	    else if(resource->kind == kCAB)
	    {
		resource->u.cab.number += elmt->amount;
	    }
	}
    }
}

/*
 * Write the code to create the jobs, resources and cabs.
 */
static void pass2(void *table_entry)
{
    struct table_entry *te;
    unsigned int i;
    
    te = (struct table_entry *)table_entry;
    
    if(te->kind == kRESOURCE)
    {
	fprintf(yyout_h, "extern KERN_resource_t *%s;\n", te->u.resource.id);
	fprintf(yyout_c, "KERN_resource_t *%s;\n", te->u.resource.id);
	fprintf(yyout_c, "U32 %s_ceiling[%u] = {", te->u.resource.id, te->u.resource.count+1);
	for(i=0; i<te->u.resource.count; i++)
	{
	    fprintf(yyout_c, "%uU, ", te->u.resource.ceiling_table[i]);
	}
	fprintf(yyout_c, "%uU};\n", te->u.resource.ceiling_table[te->u.resource.count]);
    }
    else if(te->kind == kJOB)
    {
	fprintf(yyout_c, "extern void %s(void);\n", te->u.job.entrypoint);
	fprintf(yyout_c, "KERN_job_t *%s;\n", te->u.job.id);
    }
    else if(te->kind == kCAB)
    {
	fprintf(yyout_h, "extern KERN_cab_t %s;\n", te->u.cab.id);
	fprintf(yyout_c, "KERN_cab_t %s;\n", te->u.cab.id);
	fprintf(yyout_c, "static U8 %s_buffer[%d];\n", te->u.cab.id, te->u.cab.number * (sizeof(KERN_cab_buffer_t) + te->u.cab.count - 1));
    }
}

/*
 * Write the code to call the functions which initialise jobs, CABs, and resources.
 */
static void pass3(void *table_entry)
{
    struct table_entry *te;
    
    te = (struct table_entry *)table_entry;
    if(te->kind == kJOB)
    {
      if(te->u.job.type == kPERIODIC)
	{
	  fprintf(yyout_c, "\t%s = KERN_job_create(%u, %u, %s, PERIODIC_JOB);\n", 
		  te->u.job.id, te->u.job.period, te->u.job.deadline, te->u.job.entrypoint);	
	}
      else
	{
	  fprintf(yyout_c, "\t%s = KERN_job_create(%u, %u, %s, SPORADIC_JOB);\n", 
		  te->u.job.id, te->u.job.period, te->u.job.deadline, te->u.job.entrypoint);	
	}
    }
    else if(te->kind == kRESOURCE)
    {
	fprintf(yyout_c, "\t%s = KERN_resource_create(%u, %s_ceiling);\n", 
		te->u.resource.id, te->u.resource.available, te->u.resource.id);
    }
    else if(te->kind == kCAB)
    {
	fprintf(yyout_c, "\tKERN_cab_create(&%s, %s_buffer, %u, %u);\n", 
		te->u.cab.id, te->u.cab.id, te->u.cab.number, te->u.cab.count);
    }
}


